
package com.ibm.oti.vm;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2010  All Rights Reserved
 */

import java.net.URL;
import java.net.MalformedURLException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.jar.Attributes;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.*;

public abstract class AbstractClassLoader extends ClassLoader {
	private static ClassLoader systemClassLoader;
	String[] parsedPath;
	int[] types;
	Object[] cache;
	private static class CacheLock {}
	static Object cacheLock = new CacheLock();
	FilePermission permissions[];
	private static class ManifestLock {}
	private static Object manifestLock = new ManifestLock();

public AbstractClassLoader(){
}

public AbstractClassLoader(ClassLoader p1){
	super(p1);
}

void fillCache(int i) {
	setTypeElement(i, VM.getClassPathEntryType(this, i));
	switch (types[i]) {
		case VM.CPE_TYPE_UNKNOWN:
			setCacheElement(i, cache);
			return;
		case VM.CPE_TYPE_JAR:
		case VM.CPE_TYPE_DIRECTORY:
			if (parsedPath[i] == null) {
				setParsedPathElement(i, com.ibm.oti.util.Util.toString(VM.getPathFromClassPath(i)));
			}
			File f = new File(parsedPath[i]);
			String path;
			try {
				path = f.getCanonicalPath();
			} catch (IOException e) {
				path = f.getAbsolutePath();
			}
			if (types[i] == VM.CPE_TYPE_DIRECTORY) {
				if (path.charAt(path.length() -1) != File.separatorChar) {
					StringBuilder buffer = new StringBuilder(path.length() + 1);
					path = buffer.append(path).append(File.separatorChar).toString();
				}
				setParsedPathElement(i, path);
				setCacheElement(i, cache);
			} else {
				// Must store absolute path so correct URL is created
				setParsedPathElement(i, path);
				try {
					ZipFile zf;
					zf = new JarFile(parsedPath[i]);
					setCacheElement(i, zf);
					return;
				} catch (IOException e) {}
				setTypeElement(i, VM.CPE_TYPE_UNUSABLE);
				setCacheElement(i, cache);
			}
			return;
		case VM.CPE_TYPE_TCP:
		case VM.CPE_TYPE_UNUSABLE:
			setCacheElement(i, cache);
			return;
	}
}

/**
 * Answers a stream on a resource found by looking up
 * resName using the class loader's resource lookup
 * algorithm.
 *
 * @return		InputStream
 *					a stream on the resource or null.
 * @param		resName	String
 *					the name of the resource to find.
 */
public InputStream getResourceAsStream(String resName) {
	if (resName == null || resName.length() < 1 || resName.charAt(0) == '/')
		return null;	// Do not allow absolute resource references!
	InputStream answer;
	if (this != systemClassLoader) {
		if (getParent() == null)
			answer = systemClassLoader.getResourceAsStream(resName);
		else
			answer = getParent().getResourceAsStream(resName);
		if (answer != null) return answer;
	}

	int length;
	synchronized(cacheLock) {
		length = cache.length;
	}
	for (int i = 0; i < length; ++i) {
		try {
			if (cache[i] == null) fillCache(i);
			switch (types[i]) {
				case VM.CPE_TYPE_JAR:
					ZipFile zf = (ZipFile)cache[i];
					ZipEntry entry;
					if ((entry = zf.getEntry(resName)) != null) {
						SecurityManager security = System.getSecurityManager();
						if (security != null) {
							initalizePermissions();
							if (permissions[i] == null) {
								setPermissionElement(i, new FilePermission(parsedPath[i], "read")); //$NON-NLS-1$
							}
							security.checkPermission(permissions[i]);
						}
						try {
							return zf.getInputStream(entry);
						} catch (IOException e) {
						}
					}
					continue;
				case VM.CPE_TYPE_DIRECTORY:
					StringBuilder buffer = new StringBuilder(parsedPath[i].length() + resName.length());
					String resourcePath = buffer.append(parsedPath[i]).append(resName).toString();
					InputStream result = openFile(resourcePath);
					if (result!=null)
						return result;
					continue;
			}
		} catch (SecurityException e) {}
	}
	return null;
}

private void setCacheElement(int i, Object value) {
	synchronized(cacheLock) {
		cache[i] = value;
	}
}

private void setTypeElement(int i, int value) {
	synchronized(cacheLock) {
		types[i] = value;
	}
}

private void setParsedPathElement(int i, String value) {
	synchronized(cacheLock) {
		parsedPath[i] = value;
	}
}

private void initalizePermissions() {
	synchronized(cacheLock) {
		if (permissions == null) {
			permissions = new FilePermission[cache.length];
		}
	}
}

private void setPermissionElement(int i, FilePermission value) {
	synchronized(cacheLock) {
		permissions[i] = value;
	}
}

private InputStream openFile(String resourcePath) {
	File f = new File(resourcePath);
	if (f.exists()) {
		try {
			return new BufferedInputStream(new FileInputStream(f));
		} catch (FileNotFoundException e) { }
	}
	return null;
}

protected URL findResource(final String res) {
	URL result = (URL)AccessController.doPrivileged(new PrivilegedAction() {
		public Object run() {
			for (int i=0; i<cache.length; i++) {
				URL result = findResourceImpl(i, res);
				if (result != null) return result;
			}
			return null;
		}});
	if (result != null) {
		SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			try {
				sm.checkPermission(result.openConnection().getPermission());
			} catch (IOException e) {
				return null;
			} catch (SecurityException e) {
				return null;
			}
		}
	}
	return result;
}

private URL findResourceImpl(int i, String res) {
	if (res.length() > 0 && res.charAt(0) == '/')
		return null;	// Do not allow absolute resource references!

	if (cache[i] == null) fillCache(i);
	try {
		switch (types[i]) {
			case VM.CPE_TYPE_JAR:
				ZipFile zf = (ZipFile)cache[i];
				if (zf.getEntry(res) != null)
					return new URL("jar", null, -1, toURLString(parsedPath[i] + "!/" + res), null); //$NON-NLS-1$ //$NON-NLS-2$
				return null;
			case VM.CPE_TYPE_DIRECTORY:
				StringBuilder buffer = new StringBuilder(parsedPath[i].length() + res.length());
				String resourcePath = buffer.append(parsedPath[i]).append(res).toString();
				File f = new File(resourcePath);
				if (f.exists())
					return new URL(toURLString(resourcePath));
				return null;
		}
	} catch (MalformedURLException e) {
	}
	return null;
}

protected Enumeration findResources(final String res) throws IOException {
	Vector result = (Vector)AccessController.doPrivileged(new PrivilegedAction() {
		public Object run() {
			Vector resources = new Vector();
			for (int i=0; i < cache.length; i++) {
				URL resource = findResourceImpl(i, res);
				if (resource != null) resources.addElement(resource);
			}
			return resources;
		}});
	SecurityManager sm;
	int length = result.size();
	if (length > 0 && (sm = System.getSecurityManager()) != null) {
		Vector reduced = new Vector(length);
		for (int i=0; i<length; i++) {
			URL url = (URL)result.elementAt(i);
			try {
				sm.checkPermission(url.openConnection().getPermission());
				reduced.addElement(url);
			} catch (IOException e) {
			} catch (SecurityException e) {
			}
		}
		result = reduced;
	}
	return result.elements();
}

/**
 * Answers a string representing the URL which matches the
 * given filename. The argument should be specified using the
 * standard platform rules. If it is not absolute, then it
 * is converted before use. The result will end in a slash
 * if the original path ended in a file separator.
 *
 */
static String toURLString(String filename) {
	String name = filename;
	if (File.separatorChar != '/') {
		// Must convert slashes.
		name = name.replace(File.separatorChar, '/');
	}
	name = com.ibm.oti.util.Util.urlEncode(name);
	StringBuilder buf = new StringBuilder(name.length() + 6);
	buf.append("file:"); //$NON-NLS-1$
	if (!name.startsWith("/")) { //$NON-NLS-1$
		// On Windows, absolute paths might not start with sep.
		buf.append('/');
	}
	name = buf.append(name).toString();
	return name;
}

public static void setBootstrapClassLoader(ClassLoader bootstrapClassLoader) {
	// Should only be called once
	if (systemClassLoader != null)
		throw new IllegalArgumentException();
	systemClassLoader = bootstrapClassLoader;
}

Package definePackage(String packageName, final int cacheIndex) {
	if (cacheIndex >= 0 && cache[cacheIndex] == null) {
		AccessController.doPrivileged(new PrivilegedAction() {
		public Object run() {
			fillCache(cacheIndex);
			return null;
		}});
	}

	if (cacheIndex >= 0 && types[cacheIndex] == VM.CPE_TYPE_JAR) {
		Manifest manifest = null;
		try {
			JarFile jf = (JarFile)cache[cacheIndex];
			manifest = jf.getManifest();
		} catch (IOException e) {}
		if (manifest != null) {
			synchronized(manifestLock) {
				String specTitle = null, specVersion = null, specVendor = null;
				String implTitle = null, implVersion = null, implVendor = null;
				Package result = super.getPackage(packageName);
				if (result != null) return result;
				Attributes mainAttributes = manifest.getMainAttributes();
				String value = mainAttributes.getValue(Attributes.Name.SEALED);
				boolean sealed = value != null &&
					value.toLowerCase().equals ("true"); //$NON-NLS-1$
				String dirName = packageName.replace('.', '/') + "/"; //$NON-NLS-1$
				Attributes attributes = manifest.getAttributes(dirName);
				if (attributes != null) {
					value = attributes.getValue(Attributes.Name.SEALED);
					if (value != null)
						sealed = value.toLowerCase().equals("true"); //$NON-NLS-1$
					specTitle = attributes.getValue(Attributes.Name.SPECIFICATION_TITLE);
					specVersion = attributes.getValue(Attributes.Name.SPECIFICATION_VERSION);
					specVendor = attributes.getValue(Attributes.Name.SPECIFICATION_VENDOR);
					implTitle = attributes.getValue(Attributes.Name.IMPLEMENTATION_TITLE);
					implVersion = attributes.getValue(Attributes.Name.IMPLEMENTATION_VERSION);
					implVendor = attributes.getValue(Attributes.Name.IMPLEMENTATION_VENDOR);
				}
				URL url = null;
				try {
					if (sealed) url = new URL(toURLString(parsedPath[cacheIndex]));
				} catch (MalformedURLException e) {}
				if (specTitle == null) specTitle = mainAttributes.getValue(Attributes.Name.SPECIFICATION_TITLE);
				if (specVersion == null) specVersion = mainAttributes.getValue(Attributes.Name.SPECIFICATION_VERSION);
				if (specVendor == null) specVendor = mainAttributes.getValue(Attributes.Name.SPECIFICATION_VENDOR);
				if (implTitle == null) implTitle = mainAttributes.getValue(Attributes.Name.IMPLEMENTATION_TITLE);
				if (implVersion == null) implVersion = mainAttributes.getValue(Attributes.Name.IMPLEMENTATION_VERSION);
				if (implVendor == null) implVendor = mainAttributes.getValue(Attributes.Name.IMPLEMENTATION_VENDOR);

				return definePackage(packageName,
					specTitle, specVersion, specVendor,
					implTitle, implVersion, implVendor,
					url);
			}
		}
	}
	synchronized(manifestLock) {
		Package result = super.getPackage(packageName);
		if (result != null) return result;
		return definePackage(packageName, null, null, null, null, null, null, null);
	}
}

/**
 * Answers the name of the package to which newClass belongs.
 */
String getPackageName(Class theClass)
{
	String name;
	int index;

	name = VM.getClassNameImpl(theClass);
	if((index = name.lastIndexOf('.')) == -1) return null;
	return name.substring(0, index);
}

}
